unit XLinkUtils;

// XLink Object Model 1.0.0
// Delphi 3, 4, 5, 6, and Kylix Implementation
//
// Copyright (c) 2002 by Dieter Khler
// ("http://www.philo.de/xml/")
//
// Definitions:
// - "Package" refers to the collection of files distributed by
//   the Copyright Holder, and derivatives of that collection of
//   files created through textual modification.
// - "Standard Version" refers to such a Package if it has not
//   been modified, or has been modified in accordance with the
//   wishes of the Copyright Holder.
// - "Copyright Holder" is whoever is name in the copyright or
//   copyrights for the package.
// - "You" is you, if you're thinking about copying or distributing
//   this Package.
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation
// files (the "Package"), to deal in the Package without restriction,
// including without limitation the rights to use, copy, modify,
// merge, publish, distribute, sublicense, and/or sell copies of the
// Package, and to permit persons to whom the Package is furnished
// to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Package.
//
// You may modify your copy of this Package in any way, provided
// that you insert a prominent notice in each changed file stating
// how and when you changed a file, and provided that you do at
// least one of the following:
//
// a) allow the Copyright Holder to include your modifications in
// the Standard Version of the Package.
//
// b) use the modified Package only within your corporation or
// organization.
//
// c) rename any non standard executables, units, and classes so
// the names do not conflict with standard executables, units, and
// classes, and provide a separate manual page that clearly documents
// how it differs from the standard version.
//
// d) make other distribution arrangements with the Copyright Holder.
//
// The name of the Copyright Holder may not be used to endorse or
// promote products derived from this Package without specific prior
// written permission.
//
// THE PACKAGE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// PACKAGE OR THE USE OR OTHER DEALINGS IN THE PACKAGE.

interface

uses
  XDOM_2_3,  // Extended Document Object Model 2.3.25 (or compatible)
  XLink,     // XLink Object Model 1.0.0 (or compatible)
             // Both are available at "http://www.philo.de/xml/".
  SysUtils, Classes;

type
  TXLinkView = class(TdomAbstractView)
  // Extracts all XLinks from an XDOM node tree.  The XDOM node tree is passed
  // in the constructor.  Calling UpdateXLinks() extracts the XLinks and makes
  // them available in the 'XLinks' property.  To update the 'XLinks' property
  // just call again UpdateXLinks().
  protected
    FXLinks: TXLinkCollection;
  public
    constructor Create(const doc: TdomDocument);
    destructor Destroy; override;
    procedure UpdateXLinks; virtual;
    property XLinks: TXLinkCollection read FXLinks;
  end;

  TXLinkResolver = class
  // This class is used to analyzed XLink linkbases as specified in the W3C
  // Proposed Recommendation "XML Linking Language (XLink) Version 1.0", ed.
  // Steve DeRose, Eve Maler and David Orchard, available at
  // "http://www.w3.org/TR/2000/PR-xlink-20001220/", section 5.1.5.
  //
  // Calling the analyzeLinkbase() function analyzes the local linkbase which
  // is specified in the 'url' parameter.  If the linkbase contains an absolute
  // reference to another local linkbase the other linkbase is analyzed, too.
  // This process is recusively repeated for all un-analyzed linkbases
  // referenced.  Afterwards, all XLinks found are collected in the
  // 'XLinkCollection' property.  If an error occured during evaluation the
  // analyzeLinkbase() function returns 'false', otherwise 'true' is returned.
  protected
    FAnalyzedLinkbasesSysId: TdomWideStringList;
    FXLinkCollection: TXLinkCollection;
    FDomImplementation: TdomImplementation;
    FXmlToDomParser: TXmlToDomParser;
    FXmlStandardDomReader: TXmlStandardDomReader;
    FXmlDocBuilder: TXmlDocBuilder;
  public
    constructor create; virtual;
    destructor destroy; override;
    function analyzeLinkbase(const url: wideString): boolean; virtual;
    property XLinkCollection: TXLinkCollection read FXLinkCollection;
  end;

implementation

const
  FLinkbaseArcrole: wideString = 'http://www.w3.org/1999/xlink/properties/linkbase';

// +++++++++++++++++++++++++++ TXLinkView +++++++++++++++++++++++++++
constructor TXLinkView.Create(const doc: TdomDocument);
begin
  inherited create;
  FDocument:= doc;
  FXLinks:= TXLinkCollection.Create;
end;

destructor TXLinkView.Destroy;
begin
  FXLinks.free;
  inherited Destroy;
end;

procedure TXLinkView.UpdateXLinks;
const
  prfx: wideString = 'http://www.w3.org/1999/xlink';
var
  iterator: TdomNodeIterator;
  elementToTest: TdomElement;
  typeValue,hrefValue,roleValue,arcRoleValue,titleValue,labelValue,fromValue,toValue: wideString;
  actuateValue: TXLinkActuate;
  showValue: TXLinkShow;
  lastExtended: TXLinkExtended;
  lastLocator: TXlinkLocator;
  lastArc: TXLinkArc;
  attr: TdomAttr;
begin
  lastExtended:= nil;
  lastLocator:= nil;
  lastArc:= nil;
  FXLinks.Clear;
  iterator:= FDocument.CreateNodeIterator(FDocument,[ntElement_Node],nil,true);
  try
    elementToTest:= (iterator.nextNode as TdomElement);
    while assigned(elementToTest) do begin
      with elementToTest do begin
        if attributes.namespaceAware then begin
          typeValue:= getAttributeNS(prfx,'type');
          if typeValue = 'simple' then begin
            attr:= getAttributeNodeNS(prfx,'href');
            if assigned(attr)
              then resolveRelativeUriWideStr(attr.baseUri,attr.value,hrefValue)
              else hrefValue:= '';
            roleValue:= getAttributeNS(prfx,'role');
            arcroleValue:= getAttributeNS(prfx,'arcrole');
            titleValue:= getAttributeNS(prfx,'title');
            showValue:= WideStringToShow(getAttributeNS(prfx,'show'));
            actuateValue:= WideStringToActuate(getAttributeNS(prfx,'actuate'));
            FXLinks.AppendXLinkSimple(hrefValue,roleValue,arcroleValue,titleValue,showValue,actuateValue,elementToTest);
          end else if typeValue = 'extended' then begin
            roleValue:= getAttributeNS(prfx,'role');
            titleValue:= getAttributeNS(prfx,'title');
            lastExtended:= FXLinks.AppendXLinkExtended(roleValue,titleValue,elementToTest);
          end else if typeValue = 'locator' then begin
            if assigned(lastExtended) then begin
              if lastExtended.AssociatedObject = elementToTest.parentNode then begin
                attr:= getAttributeNodeNS(prfx,'href');
                if assigned(attr)
                  then resolveRelativeUriWideStr(attr.baseUri,attr.value,hrefValue)
                  else hrefValue:= '';
                if hrefValue <> '' then begin
                  roleValue:= getAttributeNS(prfx,'role');
                  titleValue:= getAttributeNS(prfx,'title');
                  labelValue:= getAttributeNS(prfx,'label');
                  lastLocator:= lastExtended.AppendXLinkLocator(hrefValue,roleValue,titleValue,labelValue,elementToTest);
                end;
              end;
            end;
          end else if typeValue = 'arc' then begin
            if assigned(lastExtended) then begin
              if lastExtended.AssociatedObject = elementToTest.parentNode then begin
                arcroleValue:= getAttributeNS(prfx,'arcrole');
                titleValue:= getAttributeNS(prfx,'title');
                showValue:= WideStringToShow(getAttributeNS(prfx,'show'));
                actuateValue:= WideStringToActuate(getAttributeNS(prfx,'actuate'));
                fromValue:= getAttributeNS(prfx,'from');
                toValue:= getAttributeNS(prfx,'to');
                lastArc:= lastExtended.AppendXLinkArc(arcroleValue,titleValue,showValue,actuateValue,fromValue,toValue,elementToTest);
              end;
            end;
          end else if typeValue = 'resource' then begin
            if assigned(lastExtended) then begin
              if lastExtended.AssociatedObject = elementToTest.parentNode then begin
                roleValue:= getAttributeNS(prfx,'role');
                titleValue:= getAttributeNS(prfx,'title');
                labelValue:= getAttributeNS(prfx,'label');
                lastExtended.AppendXLinkResource(roleValue,titleValue,labelValue,elementToTest);
              end;
            end;
          end else if typeValue = 'title' then begin
            if assigned(lastExtended) then begin
              if lastExtended.AssociatedObject = elementToTest.parentNode
                then lastExtended.AppendXLinkTitle(elementToTest);
            end else if assigned(lastLocator) then begin
              if lastLocator.AssociatedObject = elementToTest.parentNode
                then lastLocator.AppendXLinkTitle(elementToTest);
            end else if assigned(lastArc) then begin
              if lastArc.AssociatedObject = elementToTest.parentNode
                then lastArc.AppendXLinkTitle(elementToTest);
            end;
          end;
        end;
      end;
      elementToTest:= (iterator.nextNode as TdomElement);
    end;
  finally
    iterator.detach;
    FDocument.clearInvalidNodeIterators;
  end;
end;

// ++++++++++++++++++++++++++ TXLinkResolver ++++++++++++++++++++++++++
constructor TXLinkResolver.create;
begin
  inherited create;
  FAnalyzedLinkbasesSysId:= TdomWideStringList.create;
  FXLinkCollection:= TXLinkCollection.create;
  FDomImplementation:= TdomImplementation.create(nil);
  FXmlToDomParser:= TXmlToDomParser.create(nil);
  FXmlToDomParser.DOMImpl:= FDomImplementation;
  FXmlStandardDomReader:= TXmlStandardDomReader.create(nil);
  FXmlDocBuilder:= TXmlDocBuilder.create(nil);
  FXmlStandardDomReader.NextHandler:= FXmlDocBuilder;
  FXmlStandardDomReader.PrefixMapping:= true;
  FXmlStandardDomReader.SuppressXmlns:= true;
  FXmlDocBuilder.BuildNamespaceTree:= true;
end;

destructor TXLinkResolver.destroy;
begin
  FAnalyzedLinkbasesSysId.free;
  FXLinkCollection.free;
  FDomImplementation.free;
  FXmlToDomParser.free;
  FXmlStandardDomReader.free;
  FXmlDocBuilder.free;
  inherited destroy;
end;

function TXLinkResolver.analyzeLinkbase(const url: wideString): boolean;
// Returns 'true', if the linkbase has been successfully analyzed.
// Returns 'false' otherwise.
var
  dummyDoc: TdomDocument;
  filename: TFilename;
  server,query,fragment: string;
  FLinkbaseDoc: TdomDocument;
  XLinkView: TXLinkView;
  i,j,k: integer;
  linkbaseLabel: wideString;
  linkbaseLocator: TXLinkLocator;
  ok: boolean;
begin
  result:= false;
  if FAnalyzedLinkbasesSysId.IndexOf(url) > -1
    then exit;
  FLinkbaseDoc:= nil;

  if not UriWideStrToFilename(url,filename,server,query,fragment)
    then exit;

  if not fileExists(fileName) then exit;

  FAnalyzedLinkbasesSysId.Add(url);

  try
    with FXmlToDomParser do begin
      dummyDoc:= FileToDom(Filename);
      try
        ok:= dummyDoc.validate(erReplace);
        if ok then begin
          FLinkbaseDoc:= DOMImpl.createDocument('dummy',nil);
          try
            FLinkbaseDoc.clear;
            FLinkbaseDoc.setBaseUri(url);
            FXmlDocBuilder.referenceNode:= FLinkbaseDoc;
            result:= FXmlStandardDomReader.parse(dummyDoc);
            if result then begin
              XLinkView:= TXLinkView.Create(FLinkbaseDoc);
              try
                XLinkView.UpdateXLinks;
                FXLinkCollection.AppendContentOf(XLinkView.XLinks,false);
                // evaluate new XLinks:
                with XLinkView.XLinks do begin
                  for i:= 0 to pred(Length) do begin
                    with TXLinkExtended(items[i]) do begin
                      for j:= 0 to pred(Arcs.Length) do begin
                        if TXLinkArc(Arcs.items[j]).Arcrole = FLinkbaseArcrole then begin
                          if TXLinkArc(Arcs.items[j]).Actuate = laOnLoad then begin
                            linkbaseLabel:= TXLinkArc(Arcs.items[j]).XLinkTo;
                            if linkbaseLabel = '' then begin
                              for k:= 0 to pred(Locators.Length) do begin
                                analyzeLinkbase(TXLinkLocator(Locators.items[k]).Href);
                              end; {for ...}
                              break;
                            end else begin
                              for k:= 0 to pred(Locators.Length) do begin
                                linkbaseLocator:= TXLinkLocator(Locators.items[k]);
                                if  linkbaseLocator.XLinkLabel = linkbaseLabel
                                  then analyzeLinkbase(linkbaseLocator.Href);
                              end; {for ...}
                            end; {if ... else ...}
                          end; {if ...}
                        end; {if ...}
                      end; {for ...}
                    end; {with ...}
                  end; {for ...}
                end; {with ...}
              finally
                XLinkView.Free;
              end; {try ... finally ...}
            end; {if ...}
          finally
            DOMImpl.freeDocument(FLinkbaseDoc);
          end; {try ... finally ...}
        end; {if ok ...}
      finally
        DOMImpl.freeDocument(dummyDoc);
      end; {try ... finally ...}
    end; {with ...}
  except
    result:= false;
  end;
end;

end.
